/***********************************************************************

GECCO, a platform for command and control gaming

Copyright (C) 2001 Henrik Brnhielm, Andreas Enblom, Niklas
Hallenfur, Karl Hasselstrm, Henrik Hgerstrm, Oskar Linde, Jing Fu
Zi and Jon slund

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or (at
your option) any later version.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.

***********************************************************************/

package gecco.client;

import gecco.client.animation.*;
import gecco.client.dialogs.*;
import gecco.client.infopanels.*;
import gecco.util.*;

import java.awt.*;
import java.awt.event.*;
import java.util.*;
import javax.swing.*;
import java.applet.AudioClip;
import java.net.URL;
import java.applet.Applet;

import gecco.game.dke.Def;
import java.io.FileWriter;
import java.io.File;
import java.io.OutputStreamWriter;
import java.io.IOException;
import java.io.PrintWriter;

/**
 *  This is the main game class. It contains the complete client for an
 *  arbitrary game.
 *
 * @author     Andreas Enblom
 * @created    June 27, 2001
 */
public class Game extends JFrame implements Client, Animator {
    /**
     *  The title of the game, shown as title in window.
     */
    private String gameTitle;

    /**
     *  Description of the Field
     */
    private Server server = new gecco.communication.client.ServerAmbassador(this);
    /**
     *  Description of the Field
     */
    private HostPort lastHost;
    /**
     *  Description of the Field
     */
    private Board board;
    /**
     *  Description of the Field
     */
    private Map map;
    /**
     *  Description of the Field
     */
    private PieceInfoPanel pieceInfoPanel = new PieceInfoPanel(this);

    /**
     *  The pieces that are currenlty in the game, hashed on piece handle.
     */
    private HashMap pieces;

    /**
     *  The pieces that are currenlty executing some actions. These actions are
     *  those on top of the queue and are of course not instantaneous. Hashed on
     *  action id.
     */
    private Hashtable actionExecutingPieces;

    /**
     *  The action id of the currenlty executing action for some piece. Hashed
     *  on piece handle. The inverse of actionExecutingPieces.
     */
    private Hashtable actionIds;

    /**
     *  The handle of the ghost or explosion that was added last. The handles of
     *  ghost pieces and explosion are created and maintianed by the client, and
     *  they should be negative as it is assumed that the handles that come from
     *  the server are positive.
     */
    private int lastExtraPieceId = 0;

    /**
     *  Description of the Field
     */
    private Piece selectedPiece = null;
    /**
     *  Description of the Field
     */
    private Action selectedAction = null;

    /**
     *  The listener that handles selections in the menu that displays the
     *  actions of a piece.
     */
    private ActionListener actionMenuListener;

    /**
     *  The listener that handles clicks on the board.
     */
    private MouseListener boardClicks;

    /**
     *  The listener that handles drags on the board.
     */
    private MouseMotionListener boardDrags;

    /**
     *  The menu that is displayed with piece actions, ghost pieces or ghost
     *  actions. Different from time to time.
     */
    private JPopupMenu actionMenu;

    /**
     *  The menu that contains the actions of a ghost piece (so far only remove.
     *  It is created the first time it is used, then it remains the same.
     */
    private JPopupMenu ghostActionMenu;

    /**
     *  The point that was clicked on the board to bring up the ghost menu. This
     *  point is used to place the new ghost piece where the mouse was clicked.
     */
    private Point ghostMenuClick = new Point(0, 0);

    /**
     *  All things that has to do with ghost pieces are turned on or off by this
     *  flag. Default: on.
     */
    private boolean useGhosts = false;

    // New attributes, HS.

    // Variables to play a audio clip 
    private AudioClip explosionAudioClip;
    private String audioDirName = "audio/";
    private String audioExplosionName = "Explosion.wav";
    private URL explosionAudioClipURL;

    private boolean selectedPropertiesBlocked = false;

    private Set selectedPieces = new HashSet();

    private static final double MINIMUM_PIECE_SPREAD = 3;

    private String defaultServerName = "localhost";
    private int defaultServerPort = 7000;

    // Attributes for log in Exp12
    public static final boolean LOG = ExperimentInfo.EXP12_LOG;
    String timeString;
    String logfilePrefix = "log";
    String logfileSuffix = ".txt";
    PrintWriter out;
    String logSeparator =":";


    /**
     *  Constructor for the Game object
     *
     * @param  gameTitle  Description of Parameter
     */

    // Modified HS, was: public Game(String gameTitle) {
    public Game(String gameTitle, CommandLineProperties prop){
	String serverName = prop.getStringProperty("-serverName", null);
	int serverPort = prop.getIntProperty("-serverPort", -1);
	String roleName = prop.getStringProperty("-roleName", null);

	if (prop.getBooleanProperty("-blockedProperties")) {
	    this.selectedPropertiesBlocked = true;
	}
	if (LOG) {
	    createLog();
	}
	
	////////////////////////////////////////////////////////////	

        /*
         * ------------------
         * Giant constructor.
         * ------------------
         */

        this.gameTitle = new String(gameTitle);

        // set up game control listeners

        // the behavior when an actions is selected from the menu
        // of actions for a piece.
        actionMenuListener =
            new ActionListener() {
                /**
                 *  Description of the Method
                 *
                 * @param  e  Description of Parameter
                 */
                public void actionPerformed(ActionEvent e) {
                    board.repaintBoard();
                    if (selectedPiece == null) {
                        return;
                    }
                    try {
                        ActionMenuItem item = (ActionMenuItem) e.getSource();
                        Action action = item.getMyAction();

                        // no parameter
                        if (action.getParameterType() == Action.VOID) {
			    if (selectedPieces.isEmpty()){ // Added HS 
				if (action.isInstantaneous()) {
				    server.doAction(selectedPiece.getHandle(),
						    action);
				} else {
				    selectedPiece.queue(action);
				    pieceInfoPanel.updateActionDisplay();
				    if (selectedPiece.getExecutingAction() == null) {
					executeNextAction(selectedPiece);
				    }
				}
				if (LOG) {
				    voidTypeActionLog(selectedPiece, action);
				}
				return;
			    }
			    // Added HS ///////////////////////////////////////////
			    else {
				Piece p;
				Iterator it = selectedPieces.iterator();
				while (it.hasNext()) {
				    p = (Piece) it.next();
				    if (action.isInstantaneous()) {
					server.doAction(p.getHandle(), action);
				    }
				    else {
					p.queue(action);
					if (p.getExecutingAction() == null) {
					    executeNextAction(p);
					}
				    }
				    if (LOG) {
					voidTypeActionLog(p, action);
				    }
				}
				pieceInfoPanel.updateActionDisplay();
				return;
			    }
			    ///////////////////////////////////////////////////////
                        }

                        // has parameter
                        selectAction(action);
                    } catch (Exception ex) {
                    }
                }
            };

        // The behavior when the board is clicked. Selects pieces
        // and sets up action parameters, displays menus etc.
        boardClicks =
            new MouseAdapter() {
                public void mousePressed(MouseEvent e) {
		    // Get focus to make it posible to zoom using + and -, se Board.java
		    board.requestFocusInWindow();
		    board.repaintBoard();
                    if ((e.getModifiers() & MouseEvent.BUTTON1_MASK) != 0) {
                        if (selectedAction == null) {
			    Piece p = getPieceAt(e.getPoint());
			    // Added HS ////////////////////////////////////////////////////
			    if (LOG) {
				selectPieceLog(e,p);
			    }
			    if ((e.getModifiersEx() & MouseEvent.CTRL_DOWN_MASK) != 0){
				if (selectedPiece != null) {
				    if (selectedPieces.contains(p)) {
					selectedPieces.remove(p);
					p.setSelected(false); // fr att avaktivera
					selectedPiece = null; // fr att den inte skall avaktiveras
					if (!selectedPieces.isEmpty()) {
					    p = (Piece) selectedPieces.iterator().next();
					    if (selectedPieces.size() == 1) {
						selectedPieces.clear();
					    }
					}
					else {
					    p = null;					     
					}
				    }
				    else {
					if (p != null ) {
					    if (selectedPieces.isEmpty()) {
						selectedPieces.add(selectedPiece);
					    }
					    selectedPieces.add(p);
					    selectedPiece = null; // fr att inte av aktivera nr ny adderas
					}
					else {
					    // Den valda pjsen r null, men ctrl ger ingen rensning
					    p = selectedPiece;
					}
				    }
				}
			    }
			    else {
				if (!selectedPieces.isEmpty()){
				    Iterator it = selectedPieces.iterator();
				    while (it.hasNext()) {
					Piece p2 = (Piece) it.next();
					p2.setSelected(false); // avmarkera alla pjser
				    }
				    selectedPieces.clear();
				}
			    }
			//////////////////////////////////////////////////////////////////
			    selectPiece(p);
			    return;
                       }
                        if (selectedAction.getParameterType() ==
			    Action.POSITION) {
                            Position p = board.getPositionAt(e.getPoint());
                            selectedAction.setParameter(p);
                        }
                        if ((selectedAction.getParameterType() == Action.PIECE)) {
                            Piece p = getPieceAt(e.getPoint());
                            // use only pieces from the server as parameters
                            // of actions, not ghosts or explosions
                            if ((p == null) || (p.getHandle() < 0)) {
                                unselectAction();
                                return;
                            }
			    selectedAction.setParameter(p);
                        }

			if (selectedPieces.isEmpty()){ // Added HS 
			    if (selectedAction.isInstantaneous()) {
				server.doAction(selectedPiece.getHandle(),
						selectedAction);
			    } else {
				selectedPiece.queue(selectedAction);
				
				pieceInfoPanel.updateActionDisplay();
				if (selectedPiece.getExecutingAction() == null) {
				    executeNextAction(selectedPiece);
				}
			    }
			    if (LOG) {
				actionLog(e, selectedPiece, selectedAction);
			    }
			}
			else {
			    Piece p;
			    Iterator it = selectedPieces.iterator();
			    while (it.hasNext()) {
				p = (Piece) it.next();
				if (selectedAction.isInstantaneous()) {
				    server.doAction(p.getHandle(), selectedAction);
				}
				else {
				    p.queue(selectedAction);
				    if (p.getExecutingAction() == null) {
					executeNextAction(p);
				    }
				}
				if (LOG) {
				    actionLog(e, p, selectedAction);
				}
			    } 
			    pieceInfoPanel.updateActionDisplay();
			} 
			
                        unselectAction();
                        return;
                    }
                    if ((e.getModifiers() & (MouseEvent.BUTTON2_MASK |
					     MouseEvent.BUTTON3_MASK)) != 0) {
			// Seem to only confuse, specialy with multi selsction, for removal
			// when stability and function has been verified, HS
// 			Piece p = getPieceAt(e.getPoint());
// 			if (p != null) {
// 			    selectPiece(p);
// 			}
			
			if (actionMenu != null) {
			    actionMenu.setVisible(false);
			    // Removing this and the line below fixes the resizing bug.
			    //board.remove(actionMenu);
			    actionMenu = null;
			}
			if (selectedPiece != null) {
			    if (selectedPiece instanceof Ghost) {
				actionMenu = getGhostActionMenu();
			    } else {
				actionMenu = createActionMenu(selectedPiece);
			    }
			} else if (useGhosts) {
			    ghostMenuClick = new Point(e.getPoint());
			    actionMenu = getGhostMenu();
			}
			if (actionMenu != null) {
			    // Removing this and the line above fixes the resizing bug.
			    //board.add(actionMenu);
			    actionMenu.show(board, e.getX(), e.getY());
			}
			// Added HS ////////////////////////////////////////
			if (LOG) {
			    rightClickLog(e, actionMenu);
			}
			///////////////////////////////////////////////////
		    
			return;
		    }
		}
	    };
    
        // lets the ghost pieces be draggable
        boardDrags =
            new MouseMotionAdapter() {
                /**
                 *  Description of the Method
                 *
                 * @param  e  Description of Parameter
                 */
                public void mouseDragged(MouseEvent e) {
                    if ((selectedPiece != null) &&
			(actionMenu != null) &&
			(!actionMenu.isVisible())
			&& (selectedPiece instanceof Ghost)) {
                        Position newPosition =
			    board.getPositionAt(e.getPoint());
                        selectedPiece.setPosition(newPosition);
                        board.repaintBoard();
                    }
                }
            };

        // set up game objects

        // There is a possibility of different threads trying to
        // access the thing simultaneously, this makes sure any
        // iterators over the HashMap do not throw
        // ConcurrentModificationExceptions when for example a piece
        // is removed somewhere else.
        pieces = new HashMap() {
		public Collection values() {
		    return new HashSet(super.values());
		}
	    };
	
        actionExecutingPieces = new Hashtable();
        actionIds = new Hashtable();
        board = new Board(this);
        board.addBoardMouseListener(boardClicks);
        board.addBoardMouseMotionListener(boardDrags);

        // set up layout
        getContentPane().setLayout(new BorderLayout());
        getContentPane().add(board, BorderLayout.CENTER);
        getContentPane().add(pieceInfoPanel, BorderLayout.EAST);

        // set up menus
        JMenu gameMenu = new JMenu("Game");
        JMenuItem item = new JMenuItem("Connect to game...");
        item.addActionListener(
			       new ActionListener() {
				   /**
				    *  Description of the Method
				    * || 
				    (selectedAction.getParameterType() == Action.PIECE_LIST)
				    * @param  e  Description of Parameter
				    */
				   public void actionPerformed(ActionEvent e) {
				       connectToGame();
				   }
			       });
        item.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_C,
						   ActionEvent.ALT_MASK));
        gameMenu.add(item);
        item = new JMenuItem("Disconnect");
        item.addActionListener(
			       new ActionListener() {
				   /**
				    *  Description of the Method
				    *
				    * @param  e  Description of Parameter
				    */
				   public void actionPerformed(ActionEvent e) {
				       disconnectFromGame();
				   }
			       });
        gameMenu.add(item);
        gameMenu.addSeparator();
	// Gosts are disabled in DKE
	//          item = new JCheckBoxMenuItem("Use ghost units", useGhosts);
	//          item.addItemListener(
	//              new ItemListener() {
	//                  /**
	//                   *  Description of the Method
	//                   *
	//                   * @param  e  Description of Parameter
	//                   */
	//                  public void itemStateChanged(ItemEvent e) {
	//                      if (e.getStateChange() == ItemEvent.SELECTED) {
	//                          useGhosts = true;
	//                      } else {
	//                          useGhosts = false;
	//                      }
	//                  }
	//              });
	//          gameMenu.add(item);

	// New menu item, HS. Disabled in experiment
// 	item = new JCheckBoxMenuItem("Block some properies", selectedPropertiesBlocked);
// 	item.addItemListener(
// 			     new ItemListener() {
// 				 public void itemStateChanged(ItemEvent e) {
// 				     if (e.getStateChange() == ItemEvent.SELECTED) {
// 					 selectedPropertiesBlocked = true;
// 				     } else {
// 					 selectedPropertiesBlocked = false;
// 				     }
// 				 }
// 			     });
// 	gameMenu.add(item);

	gameMenu.addSeparator();

        item = new JMenuItem("Quit");
        item.addActionListener(
			       new ActionListener() {
				   /**
				    *  Description of the Method
				    *
				    * @param  e  Description of Parameter
				    */
				   public void actionPerformed(ActionEvent e) {
				       quit();
				   }
			       });
        item.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_Q,
						   ActionEvent.ALT_MASK));
        gameMenu.add(item);

        JMenu infoMenu = new JMenu("Info");
        item = new JMenuItem("About...");
        item.addActionListener(
			       new ActionListener() {
				   /**
				    *  Description of the Method
				    *
				    * @param  e  Description of Parameter
				    */
				   public void actionPerformed(ActionEvent e) {
				       About.display(Game.this);
				   }
			       });
        item.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_I,
						   ActionEvent.ALT_MASK));
        infoMenu.add(item);

        JMenuBar menubar = new JMenuBar();
        menubar.add(gameMenu);
        menubar.add(infoMenu);
        setJMenuBar(menubar);

        // set up window
        setTitle(gameTitle);
        pack();
        Dimension screenSize = Toolkit.getDefaultToolkit().getScreenSize();
	//          setSize(screenSize.width * 3 / 4, screenSize.height * 2 / 3);
	//          setLocation(screenSize.width / 8, screenSize.height / 6);

	// Full screen
	//  	setSize(screenSize.width, screenSize.height);
	//  	setLocation(0, 0);

	int screenWidth = screenSize.width;
	int screenHeight = screenSize.height;
	if ((screenWidth > 1280) && (screenHeight > 1024)) {
	    setSize(1280,1024);
	}
	else {
	    // Maximize frame
	    setExtendedState(MAXIMIZED_BOTH);
	}

        addWindowListener(
			  new WindowAdapter() {
			      /**
			       *  Description of the Method
			       *
			       * @param  e  Description of Parameter
			       */
			      public void windowClosing(WindowEvent e) {
				  quit();
			      }
			  });
	initAudio();
	setVisible(true);

	// Added HS ///////////////////////////////////////////////////////////////////
	// Make it posible to zoom by typing num + and - or A resp. Z
	class ZoomInAction extends AbstractAction{
	    public void actionPerformed(ActionEvent e){
		board.zoomIn();
	    }
	};
	class ZoomOutAction extends AbstractAction{
	    public void actionPerformed(ActionEvent e){
		board.zoomOut();
	    }
	};
	ZoomInAction zoomInAction = new ZoomInAction();
	ZoomOutAction zoomOutAction = new ZoomOutAction();
	InputMap im = getRootPane().getInputMap(getRootPane().WHEN_ANCESTOR_OF_FOCUSED_COMPONENT);
	ActionMap am = getRootPane().getActionMap();
	im.put(KeyStroke.getKeyStroke(KeyEvent.VK_A, 0), "zoomIn");
	im.put(KeyStroke.getKeyStroke(KeyEvent.VK_PLUS, 0), "zoomIn");
	im.put(KeyStroke.getKeyStroke(KeyEvent.VK_ADD, 0), "zoomIn");
	im.put(KeyStroke.getKeyStroke(KeyEvent.VK_Z, 0), "zoomOut");
	im.put(KeyStroke.getKeyStroke(KeyEvent.VK_MINUS, 0), "zoomOut");
	im.put(KeyStroke.getKeyStroke(KeyEvent.VK_SUBTRACT, 0), "zoomOut");
	am.put("zoomIn", zoomInAction);
	am.put("zoomOut", zoomOutAction);
	///////////////////////////////////////////////////////////////////////////////
	connectToGame(serverName, serverPort, roleName);
    }

    private void initAudio(){
 	try {
	    explosionAudioClipURL = Game.class.getResource(audioDirName + audioExplosionName);
	    if (explosionAudioClipURL == null) {
		// Could not find audio file.
		System.err.println("Missing file: " + audioDirName + audioExplosionName );
		explosionAudioClip = null;
		return;
	    }
	    explosionAudioClip = Applet.newAudioClip(explosionAudioClipURL);
	    if (explosionAudioClip == null) {
		// Could not load audio file.
		System.err.println("Could not load audio file: URL = " + explosionAudioClipURL);
		return;
	    }
	    
	} catch (Exception e) {
	    e.printStackTrace();
	    explosionAudioClip = null;
	} 
    }

    protected void playAudioClip(AudioClip clip){
	if (clip != null) {
	    clip.play();
	}
    }
	


    /**
     *  Returns the piece at the given point of the board, or null if no such
     *  piece exists.
     *
     * @param  point  Description of Parameter
     * @return        The PieceAt value
     */
    public Piece getPieceAt(Point point) {
	Iterator iter;
	synchronized (pieces) {
	    iter = pieces.values().iterator();
	}
        Piece piece = null;
        while (iter.hasNext()) {
            Piece p = (Piece) iter.next();
            Rectangle r = board.getPieceBounds(p);
            if (r.contains(point)) {
                piece = p;
            }
        }
        return piece;
    }


    /**
     *  Returns all pieces of the game that intersect the given rectangle. The
     *  rectangle is in actual screen coordinates, relative to the component of
     *  the board where the map and the pieces are painted.
     *
     * @param  rect  Description of Parameter
     * @return       The PiecesIntersecting value
     */
    public Enumeration getPiecesIntersecting(Rectangle rect) {
	Iterator all;
	synchronized (pieces) {
	    all = pieces.values().iterator();
	}
        Vector intersecting = new Vector();
        while (all.hasNext()) {
            Piece p = (Piece) all.next();
            Rectangle r = board.getPieceBounds(p);
            if (r.intersects(rect)) {
                intersecting.addElement(p);
            }
        }
        return intersecting.elements();
    }


    /**
     *  Returns the preferred center of the map. This is the center of the
     *  current screen or, if there is a selected piece, the position of the
     *  selected piece. The preferred center is the center around which the
     *  board will zoom in and out.
     *
     * @return    The preferred center of the map.
     */
    public Position getPreferredCenter() {
        if (selectedPiece == null) {
            Dimension boardSize = board.boardSize();
            return board.getPositionAt(new Point(boardSize.width / 2,
						 boardSize.height / 2));
        } else {
            return selectedPiece.getPosition();
        }
    }


    /**
     *  Returns the positions of the queued actions (and the currently executing
     *  action) of the selected piece, or NULL if no piece is selected. The
     *  position of an action is the position of its argument (if any). The
     *  collection points is started with the position of the selected piece.
     *
     * @return    The SelectedPiecePositions value
     */
    public Enumeration getSelectedPiecePositions() {
	return getPiecePositions(selectedPiece);
    }

    public Enumeration getPiecePositions(Piece piece){
	if (piece == null) {
            return null;
        }
        Vector positions = new Vector();

        // first, the position of the piece.
        positions.add(piece.getPosition());

        // executing action
        Action ea = piece.getExecutingAction();
        if (ea != null) {
            if (ea.getParameterType() == Action.POSITION) {
                Position p = (Position) ea.getParameter();
                positions.add(new Position(p));
            } else if (ea.getParameterType() == Action.PIECE) {
                Piece p = (Piece) ea.getParameter();
                positions.add(new Position(p.getPosition()));
            }
        }

        // queued actions
        Enumeration actions = piece.getAllQueuedActions().elements();
        while (actions.hasMoreElements()) {
            Action a = (Action) actions.nextElement();
            if (a.getParameterType() == Action.POSITION) {
                Position p = (Position) a.getParameter();
                positions.add(new Position(p));
            } else if (a.getParameterType() == Action.PIECE) {
                Piece p = (Piece) a.getParameter();
                positions.add(new Position(p.getPosition()));
            }
        }
        return positions.elements();
    }

    /**
     * Get the SelectedPropertiesBlocked value.
     * @return the SelectedPropertiesBlocked value.
     */
    public boolean isSelectedPropertiesBlocked() {
	return selectedPropertiesBlocked;
    }


    /*
     * --------------------------------------------------------
     * Client interface methods. For comments, see Client.java.
     * --------------------------------------------------------
     */

    /**
     *  Description of the Method
     *
     * @param  points  Description of Parameter
     * @param  colors  Description of Parameter
     */
    public void mapUpdate(Point[] points, int[] colors) {
        int size = Math.min(points.length, colors.length);
        for (int i = 0; i < size; ++i) {
            board.mapUpdate(points[i], colors[i]);
        }
        board.repaintBoard();
    }


    /**
     *  Description of the Method
     *
     * @param  p      Description of Parameter
     * @param  color  Description of Parameter
     */
    public void mapUpdate(Point p, int color) {
        board.mapUpdate(p, color);
        board.repaintBoard();
    }


    /**
     *  Makes a number of automatons invisible.
     *
     * @param  points  A vector of automaton coordinates.
     */
    public void hideMapPoints(Point[] points) {
        for (int i = 0; i < points.length; ++i) {
            board.hideMapPoint(points[i]);
        }
        board.repaintBoard();
    }


    /**
     *  Makes an automaton in the map invisible.
     *
     * @param  p  The coordinate of the automaton.
     */
    public void hideMapPoint(Point p) {
        board.hideMapPoint(p);
        board.repaintBoard();
    }


    /**
     *  Description of the Method
     *
     * @param  handle   Description of Parameter
     * @param  message  Description of Parameter
     */
    public void unitMessage(int handle, String message) {
	// Added HS ///////////////////////////////////////////////////////////////
	if (message.equals("Action aborted!")) { // filtrera bort dessa meddelanden
	    return;
	}
	else if (selectedPropertiesBlocked){ // filtrera bort om blockerade properties
	    if (message.startsWith(Def.getCombatEndedString())){
		return;
	    }
	}
	
	/////////////////////////////////////////////////////////////////////////////
	Piece p;
	synchronized (pieces) {
	    p = (Piece) pieces.get(new Integer(handle));
	}
        if (p == null) {
            return;
        }
        p.addMessage(message);
        if (p == selectedPiece) {
            pieceInfoPanel.updateMessageDisplay();
        }
	// Added HS ////////////////////////////////////////////////////////////////////////////////
	if (message.startsWith(Def.getWarningString())) {
	    String warning = p.getName() + ": " + message.substring(Def.getWarningString().length());
	    pieceInfoPanel.warningMessage(warning);
	}
	/////////////////////////////////////////////////////////////////////////////////////////////
    }


    /**
     *  Description of the Method
     *
     * @param  time  Description of Parameter
     */
    public void timeUpdate(String time) {
	if (LOG) {
	    timeString = time;
	}
	pieceInfoPanel.updateTime(time);
    }

    public void warningMessage(String message){
	pieceInfoPanel.warningMessage(message);
    }


    /**
     *  Adds the specified piece, as described in Client.java. If there is a
     *  ghost piece with the same type and position, that ghost piece will be
     *  removed and replaced with the new piece. If the ghost is selected, the
     *  selection will carry over to the new piece.
     *
     * @param  handle      The feature to be added to the Piece attribute
     * @param  name        The feature to be added to the Piece attribute
     * @param  pos         The feature to be added to the Piece attribute
     * @param  actions     The feature to be added to the Piece attribute
     * @param  properties  The feature to be added to the Piece attribute
     * @param  type        The feature to be added to the Piece attribute
     */
    public void addPiece(int handle, String name, Position pos,
			 Set actions, java.util.Map properties, String type) {
        Piece p = new Piece(handle, pos, name, type);
	synchronized (pieces) {
	    pieces.put(new Integer(p.getHandle()), p);
	}
        if (actions != null) {
	    Iterator iter = actions.iterator();
            while (iter.hasNext()) {
                p.addAction((Action) iter.next());
            }
        }
        if (properties != null) {
	    Iterator iter = properties.keySet().iterator();
            while (iter.hasNext()) {
		String key = (String) iter.next();
		String value = (String) properties.get(key);
                p.setProperty(key, value);
            }
        }

        // find any ghost piece at the same location and with the same type
	Iterator iter;
	synchronized (pieces) {
	    iter = pieces.values().iterator();
	}
        while (iter.hasNext()) {
            Piece q = (Piece) iter.next();
            if ((q.getPosition().equals(pos))
		&& (q instanceof Ghost)
		&& (q.getType().equals(type))) {
                if (selectedPiece == q) {
                    selectPiece(p);
                }
		synchronized (pieces) {
		    pieces.remove(q);
		}
            }
        }

        board.repaintBoard();
    }


    /**
     *  Description of the Method
     *
     * @param  handle         Description of Parameter
     * @param  newProperties  Description of Parameter
     */
    public void updatePiece(int handle, Property[] newProperties) {
	Piece p;
	synchronized (pieces) {
	    p = (Piece) pieces.get(new Integer(handle));
	}
        if (p == null) {
            return;
        }
        for (int i = 0; i < newProperties.length; ++i) {
	    // Modified HS //////////////////////////////
	    if (newProperties[i].getValue() == null) {
		p.removeProperty(newProperties[i]);
	    }
	    else {
		/////////////////////////////////////////
		p.setProperty(newProperties[i]);
	    }
        }
        if (p == selectedPiece) {
            pieceInfoPanel.updatePropertyDisplay();
        }
    }


    /**
     *  Description of the Method
     *
     * @param  handle       Description of Parameter
     * @param  newPosition  Description of Parameter
     */
    public void updatePiece(int handle, Position newPosition) {
	Piece p;
	synchronized (pieces) {
	    p = (Piece) pieces.get(new Integer(handle));
	}
        if (p == null) {
            return;
        }
	assureMinimumSpread(p, newPosition); // Added HS
        p.setPosition(newPosition);
        board.repaintBoard();
    }

    // Added HS /////////////////////////////////////////////////////
    public void assureMinimumSpread(Piece p, Position pos){
	Iterator it = pieces.values().iterator();
	double spread;
	double dx;
	double dy;
	Piece p2;
	while (it.hasNext()) {
	    p2 = (Piece) it.next();
	    dx = p2.getPosition().getX()-pos.getX();
	    dy = p2.getPosition().getY()-pos.getY();
	    spread = Math.sqrt(dx*dx+dy*dy);
	    if (p != p2 && spread < MINIMUM_PIECE_SPREAD) {
		double cosValue;
		double sinValue;
		if (spread == 0) {
		    double randomAngle = Math.random()*2*Math.PI;
		    cosValue = Math.cos(randomAngle);
		    sinValue = Math.sin(randomAngle);
		}
		else {
		    cosValue = dx / spread;
		    sinValue = dy / spread;
		}
		double addOn = spread - MINIMUM_PIECE_SPREAD;
		pos.translate(addOn*cosValue, addOn*sinValue);
	    }
	}
    }
    private void createLog(){
	try {
	    int nr = 0;
	    File f = new File(logfilePrefix+nr+logfileSuffix);
	    while (f.exists()) {
		nr++;
		f = new File(logfilePrefix+nr+logfileSuffix);
	    }
	    f.createNewFile();
	    out = new PrintWriter(new FileWriter(f));
	} catch (Exception e) {
	    System.err.println(e.getMessage());
	    e.printStackTrace();
	    System.exit(1);
	}
    }
    private int timeStringToIntSeconds(String timeString){
	int seconds;
	int timeBeginCharacter = ':';
	int yearsEndCharacter = 'y';
	int daysEndCharacter = 'd';
	int hoursEndCharacter = 'h';
	int minutesEndCharacter = 'm';
	int secondsEndCharacter = 's';

	int indexOfTimeBeginCharacter = timeString.indexOf(timeBeginCharacter);
	timeString = timeString.substring(indexOfTimeBeginCharacter + 1);
	int indexOfYearsEndCharacter = timeString.indexOf(yearsEndCharacter);
	String yearsAsString = timeString.substring(0, indexOfYearsEndCharacter);
	timeString = timeString.substring(indexOfYearsEndCharacter + 1);
	int indexOfDaysEndCharacter = timeString.indexOf(daysEndCharacter);
	String daysAsString = timeString.substring(0, indexOfDaysEndCharacter);
	timeString = timeString.substring(indexOfDaysEndCharacter + 1);
	int indexOfHoursEndCharacter = timeString.indexOf(hoursEndCharacter);
	String hoursAsString = timeString.substring(0, indexOfHoursEndCharacter);
	timeString = timeString.substring(indexOfHoursEndCharacter + 1);
	int indexOfMinutesEndCharacter = timeString.indexOf(minutesEndCharacter);
	String minutesAsString = timeString.substring(0, indexOfMinutesEndCharacter);
	timeString = timeString.substring(indexOfMinutesEndCharacter + 1);
	int indexOfSecondsEndCharacter = timeString.indexOf(secondsEndCharacter);
	String secondsAsString = timeString.substring(0, indexOfSecondsEndCharacter);

	int yearsAsInt = Integer.parseInt(yearsAsString);
	int daysAsInt = Integer.parseInt(daysAsString);
	int hoursAsInt = Integer.parseInt(hoursAsString);
	int minutesAsInt = Integer.parseInt(minutesAsString);
	int secondsAsInt = Integer.parseInt(secondsAsString);

	seconds = secondsAsInt + minutesAsInt*60 + hoursAsInt*60*60 
	    + daysAsInt*24*60*60 + yearsAsInt*365*24*60*60;
	return seconds;
    }
    private void printTimeInfo(){
	int timeInSeconds = timeStringToIntSeconds(timeString);
	out.print(timeInSeconds+logSeparator);
    }
    private void printPieceInfo(Piece p){
	if (p == null) {
	    out.print("null"+logSeparator);
	}
	else {
	    out.print(p.getName()+logSeparator);
	    out.print(p.getHandle()+logSeparator);
	}
    }
    private void printActionInfo(Action a){
	if (a == null){
	    out.print("null"+logSeparator);
	}
	else {
	    out.print(a.getName()+logSeparator);
	    int parameterType = a.getParameterType();
	    if (parameterType == Action.PIECE) {
		Piece piece = (Piece) a.getParameter();
		out.print("PIECE"+logSeparator);
		out.print(piece.getName()+logSeparator);
		out.print(piece.getHandle()+logSeparator);
	    }
	    else if (parameterType == Action.POSITION) {
		Position position = (Position) a.getParameter();
		out.print("POSITION"+logSeparator);
		out.print(position.getX()+","+position.getY()+logSeparator);
	    }
	    else {
		out.print("VOID"+logSeparator); 
	    }
	}
    }
    private void printMouseEventInfo(MouseEvent e){
	out.print(e.getX()+","+e.getY()+logSeparator);
	out.print((e.isControlDown()?"yes":"no")+logSeparator);
    }
    private void endRecord(){
	out.println();
	out.flush();
    }
    private void selectPieceLog(MouseEvent e, Piece p){
	printTimeInfo();
	out.print("MousePressed"+logSeparator);
	out.print("Left "+logSeparator);
	printMouseEventInfo(e);
	out.print("Piece Selected"+logSeparator);
	printPieceInfo(p);
	endRecord();
    }
    private void rightClickLog(MouseEvent e, JPopupMenu actionMenu){
	printTimeInfo();
	out.print("MousePressed"+logSeparator);
	out.print("Right"+logSeparator);
	printMouseEventInfo(e);
	out.print((actionMenu==null? "no":"yes")+logSeparator);
	endRecord();
    }
    private void actionLog(MouseEvent e, Piece p, Action a){
	printTimeInfo();
	out.print("MousePressed"+logSeparator);
	out.print("Left "+logSeparator);
	printMouseEventInfo(e);
	out.print("Action selected"+logSeparator);
	printPieceInfo(p);
	printActionInfo(a);
	endRecord();
    }
    private void voidTypeActionLog(Piece p, Action a){
	printTimeInfo();
	out.print("VOID action selected"+logSeparator);
	printPieceInfo(p);
	printActionInfo(a);
	endRecord();
    }
    private void abortExecutingActionLog(Action a){
	printTimeInfo();
	out.print("Abort executing action"+logSeparator);
	printActionInfo(a);
	endRecord();
    }	
    // Anvnds av PropertyDisplay
    public void pieceInfoLog(Piece p){
	printTimeInfo();
	out.print("Piece info selected"+logSeparator);
	printPieceInfo(p);
	endRecord();
    }
    // Anvnds av ActionDisplay.removeSelectedActions()
    public void abortQueuedActionLog(Action a){
	printTimeInfo();
	out.print("Abort queued action"+logSeparator);
	printActionInfo(a);
	endRecord();
    }
    // Anvnds av Board.zoomIn / zoomOut
    public void zoomInfoLog(String direction){
	printTimeInfo();
	out.print("Zoom" +logSeparator);
	out.print(direction + logSeparator);
	endRecord();
    }
    /////////////////////////////////////////////////////////////////


    /**
     *  Description of the Method
     *
     * @param  handle  Description of Parameter
     */
    public void removePiece(int handle) {
	Piece p;
	synchronized (pieces) {
	    p = (Piece) pieces.get(new Integer(handle));
	}
        if (p == null) {
            return;
        }
        Piece ghost = addGhostPiece(p.getPosition(), p.getType());
        if (selectedPiece == p) {
            selectPiece(ghost);
        }
        destroyPiece(handle, false);
        board.repaintBoard();
    }


    /**
     *  Description of the Method
     *
     * @param  handle  Description of Parameter
     */
    public void destroyPiece(int handle) {
        destroyPiece(handle, true);
    }


    /**
     *  Adds a feature to the PieceAction attribute of the Game object
     *
     * @param  handle  The feature to be added to the PieceAction attribute
     * @param  a       The feature to be added to the PieceAction attribute
     */
    public void addPieceAction(int handle, Action a) {
	Piece p;
	synchronized (pieces) {
	    p = (Piece) pieces.get(new Integer(handle));
	}
        if (p == null) {
            return;
        }
        p.addAction(a);
    }


    /**
     *  Description of the Method
     *
     * @param  handle  Description of Parameter
     * @param  a       Description of Parameter
     */
    public void removePieceAction(int handle, Action a) {
	Piece p;
	synchronized (pieces) {
	    p = (Piece) pieces.get(new Integer(handle));
	}
        if (p == null) {
            return;
        }
        p.removeAction(a);
    }


    /**
     *  Description of the Method
     *
     * @param  id       Description of Parameter
     * @param  ok       Description of Parameter
     * @param  message  Description of Parameter
     */
    public void actionReply(long id, boolean ok) {
	Piece p;
	synchronized (pieces) {
	    p = (Piece) actionExecutingPieces.get(new Long(id));
	}
        if (p == null) {
            // instantaneous actions are caught here
            return;
        }
        actionExecutingPieces.remove(new Long(id));
        actionIds.remove(new Integer(p.getHandle()));
        if (ok) {
            executeNextAction(p);
        } else {
            p.abortAllExecutingActions();
        }
        if (p == selectedPiece) {
            pieceInfoPanel.updateActionDisplay();
        }
        board.repaintBoard();
    }


    /**
     *  Description of the Method
     */
    public void connectionLost() {
        Toolkit.getDefaultToolkit().beep();
        gecco.client.dialogs.Error.display("Connection to server lost. "
					   + "Please try to connect again.", this);
        disconnectFromGame();
    }


    /**
     *  Description of the Method
     *
     * @param  progress  Description of Parameter
     */
    public void mapLoadingProgressUpdate(double progress) {
        LoadingInfo.update(progress);
    }


    public void abortAllActions(int pieceHandle){
	// Debug info
	System.out.println("Game: abortAllActions(" + pieceHandle + ")");
	
	Piece p = (Piece)pieces.get(new Integer(pieceHandle));
	abortExecutingAction(p, true);
    }


    /*
     * -----------------------------------------------------------
     * Animator interface methods. For comments, see Animator.java
     * -----------------------------------------------------------
     */

    /**
     *  Description of the Method
     *
     * @param  p  Description of Parameter
     */
    public void repaintAnimation(AnimatedPiece p) {
        board.repaintBoard();
    }


    /**
     *  Description of the Method
     *
     * @param  p  Description of Parameter
     */
    public void animationDone(AnimatedPiece p) {
        destroyPiece(p.getHandle(), false);
    }


    /*
     * ------------------------------
     * External game control methods.
     * ------------------------------
     */

    /**
     *  Aborts the currently executing action of the given piece. Optionally,
     *  all queued actions are aborted, if not the next action in queue is
     *  executed.
     *
     * @param  p         The piece.
     * @param  abortAll  Indiactes whether or not all executing actions should
     *      be aborted.
     */
    public void abortExecutingAction(Piece p, boolean abortAll) {
        if (p == null) {
            return;
        }
        Long id = (Long) actionIds.get(new Integer(p.getHandle()));
        if (id == null) {
            // piece has executing action
            return;
        }
        server.abortAction(id.longValue());
        actionExecutingPieces.remove(id);
        actionIds.remove(new Integer(p.getHandle()));
        if (abortAll) {
            p.abortAllExecutingActions();
        } else {
            executeNextAction(p);
        }
        if (p == selectedPiece) {
            pieceInfoPanel.updateActionDisplay();
        }
        board.repaintBoard();
    }

    /**
     * <code>abortExecutingAction</code> support method for multi selection of pieces
     *
     * @param s a <code>Set</code> containing <code>Piece</code>
     */
    public void abortExecutingAction(Set s){
	Piece p;
	Iterator it = s.iterator();
	while (it.hasNext()) {
	    p = (Piece) it.next();
	    abortExecutingAction(p);
	}
    }
	


    /**
     *  Aborts the currently executing action of the given piece. The next
     *  action in queue is then executed.
     *
     * @param  p  Description of Parameter
     */
    public void abortExecutingAction(Piece p) {
	if (LOG) {
	    abortExecutingActionLog(p.getExecutingAction());
	}
        abortExecutingAction(p, false);
    }

    public void gameOver(String message){
	//Debug info
	System.out.println(message);
	String fireworkGIF = "icons/fyrverkeri.gif";
	URL fireworkURL = getClass().getResource(fireworkGIF);
	Icon fireworkIcon = new ImageIcon(fireworkURL);
	JOptionPane.showMessageDialog(this, message, "GAME OVER!",
				      JOptionPane.INFORMATION_MESSAGE, fireworkIcon);
    }

    public void coffeeBreak(String message){
	//Debug info
	System.out.println("CoffeeBreak : " + message);
	board.setVisible(false);
	String coffeeGIF = "icons/kaffe.gif";
	URL coffeeURL = getClass().getResource(coffeeGIF);
	Icon coffeeIcon = new ImageIcon(coffeeURL);
	JOptionPane.showMessageDialog(this, message, "Coffee Break",
				      JOptionPane.INFORMATION_MESSAGE, coffeeIcon);
	board.setVisible(true);
    }


    /**
     *  Returns the menu that displays all ghost pieces that are available. Sets
     *  up the behavior so that the selected ghost is placed on the board.
     *
     * @return    The GhostMenu value
     */
    private JPopupMenu getGhostMenu() {
        JPopupMenu ghostMenu = new JPopupMenu();
        Enumeration e = Symbols.getAllPieceTypes();
        while (e.hasMoreElements()) {
            String type = (String) e.nextElement();
            Image symbol = Symbols.getPieceSymbol(type);
            JMenuItem item = new JMenuItem(type, new ImageIcon(symbol));
            item.setActionCommand(type);
            item.addActionListener(new ActionListener() {
		    public void actionPerformed(ActionEvent e) {
			String pieceType = e.getActionCommand();
			Position position =
			    board.getPositionAt(ghostMenuClick);
			addGhostPiece(position, pieceType);
		    }
		});
            ghostMenu.add(item);
        }
        return ghostMenu;
    }


    /**
     *  Returns the menu that contains actions for the ghosts.
     *
     * @return    The GhostActionMenu value
     */
    private JPopupMenu getGhostActionMenu() {
        if (ghostActionMenu != null) {
            return ghostActionMenu;
        }
        ghostActionMenu = new JPopupMenu();
        JMenuItem item = new JMenuItem("Remove ghost");
        item.addActionListener(
			       new ActionListener() {
				   /**
				    *  Description of the Method
				    *
				    * @param  e  Description of Parameter
				    */
				   public void actionPerformed(ActionEvent e) {
				       if ((selectedPiece != null) &&
					   (selectedPiece instanceof Ghost)) {
					   destroyPiece(selectedPiece.getHandle(),
							false);
				       }
				   }
			       });
        ghostActionMenu.add(item);
        return ghostActionMenu;
    }


    /*
     * ------------------------------
     * Internal game control methods.
     * ------------------------------
     */

    /**
     *  Description of the Method
     *
     * @param  p  Description of Parameter
     */
    private void executeNextAction(Piece p) {
        if (p == null) {
            return;
        }
        Action nextAction = p.executeNextAction();
        if (nextAction == null) {
            return;
        }
        long nextId = server.doAction(p.getHandle(), nextAction);
        actionExecutingPieces.put(new Long(nextId), p);
        actionIds.put(new Integer(p.getHandle()), new Long(nextId));
        if (selectedPiece == p) {
            pieceInfoPanel.updateActionDisplay();
        }
    }


    /**
     *  Description of the Method
     *
     * @param  p  Description of Parameter
     */
    private void selectPiece(Piece p) {
        if (selectedPiece != null) {
            selectedPiece.setSelected(false);
        }
        if (p != null) {
            p.setSelected(true);
        }
        selectedPiece = p;
        unselectAction();
        pieceInfoPanel.setPiece(selectedPiece);
        board.repaintBoard();
    }


    /**
     *  Description of the Method
     */
    private void unselectPiece() {
        selectPiece(null);
    }


    /**
     *  Description of the Method
     *
     * @param  a  Description of Parameter
     */
    private void selectAction(Action a) {
        selectedAction = a;
        board.setBoardCursor(new Cursor(Cursor.CROSSHAIR_CURSOR));
        if (a == null) {
            board.setBoardCursor(new Cursor(Cursor.DEFAULT_CURSOR));
        }
    }


    /**
     *  Description of the Method
     */
    private void unselectAction() {
        selectAction(null);
    }


    /**
     *  Removes the piece with the given handle (if any) from the game.
     *
     * @param  animateExplosion  Should there be an explosion?
     * @param  handle            Description of Parameter
     */
    private void destroyPiece(int handle, boolean animateExplosion) {
	Piece p;
	synchronized (pieces) {
	    p = (Piece) pieces.remove(new Integer(handle));
	}
        if (p == null) {
            return;
        }
        if (animateExplosion) {
            --lastExtraPieceId;
            Explosion explosion = new Explosion(lastExtraPieceId,
						p.getPosition(),
						p.getName(), p.getType(),
						this);
	    synchronized (pieces) {
		pieces.put(new Integer(explosion.getHandle()), explosion);
	    }
	    // play explosion sound
	    playAudioClip(explosionAudioClip);
            explosion.startAnimation();
        }
        if (selectedPiece == p) {
            unselectPiece();
        }
        board.repaintBoard();
    }


    /**
     *  Adds a ghost piece with the specified properties. The newly created
     *  ghost symbol is returned. If there is already a ghost piece with the
     *  same type at the exact same location, no ghost will be added. In the
     *  latter case, the existing ghost piece is returned. In the ghost
     *  functionality is disabled (useGhosts flag), nothing is done and null is
     *  returned.
     *
     * @param  pos   The feature to be added to the GhostPiece attribute
     * @param  type  The feature to be added to the GhostPiece attribute
     * @return       Description of the Returned Value
     */
    private Piece addGhostPiece(Position pos, String type) {
        if (!useGhosts) {
            return null;
        }

        // make sure there is no already a similar ghost piece there
        Iterator iter;
	synchronized (pieces) {
	    iter = pieces.values().iterator();
	}
        while (iter.hasNext()) {
            Piece p = (Piece) iter.next();
            if ((p.getPosition().equals(pos))
		&& (p instanceof Ghost)
		&& (p.getType().equals(type))) {
                return p;
            }
        }

        // create a new ghost piece and put it there
        --lastExtraPieceId;
        Ghost ghost = new Ghost(lastExtraPieceId, pos, "Ghost", type);
	synchronized (pieces) {
	    pieces.put(new Integer(ghost.getHandle()), ghost);
	}
        board.repaintBoard();
        return ghost;
    }


    /**
     *  Creates the menu that displays the actions of a piece. The appropriate
     *  behavior of the menu is set up.
     *
     * @param  p  Description of Parameter
     * @return    Description of the Returned Value
     */
    // Modified HS ///////////////////////////////////////////////////
    private JPopupMenu createActionMenu(Piece p) {
	unselectAction();
        JPopupMenu menu = new JPopupMenu();
        Vector actionMenuItems = new Vector();

	if (selectedPieces.isEmpty()){
	    Enumeration actions = p.getActions();
	    while (actions.hasMoreElements()) {
		Action a = (Action) actions.nextElement();
		ActionMenuItem item = new ActionMenuItem(a);
		item.addActionListener(actionMenuListener);
		actionMenuItems.add(item);
	    }
	}
	else {
	    // skapa ett set fr varje pjs, ladda in actions fr varje pjs
	    // behll endast de actions som finns i alla set, retainAll(Collection c)
	    // gr menyn frn kvarstende actions
	    Set[] act = new Set[selectedPieces.size()];
	    int i = 0;
	    Iterator it = selectedPieces.iterator();
	    while (it.hasNext()) {
		Set set = new TreeSet();
		Piece p2 = (Piece) it.next();
		Enumeration actions = p2.getActions();
		while (actions.hasMoreElements()) {
		    Action a = (Action) actions.nextElement();
		    set.add(a);
		}
		act[i] = set;
		i++;
	    }
	    for (int j = 1; j < i; j++) {
		act[0].retainAll(act[j]);
	    }
	    it = act[0].iterator();
	    while (it.hasNext()) {
		Action a = (Action) it.next();
		ActionMenuItem item = new ActionMenuItem(a);
		item.addActionListener(actionMenuListener);
		actionMenuItems.add(item);
	    }
	}
	if (actionMenuItems.size() <= 0) {
	    JMenuItem noActions = new JMenuItem("No actions available");
	    noActions.setEnabled(false);
	    menu.add(noActions);
	} else {
	    JMenuItem stopAction = new JMenuItem("Stop ongoing action");
	    if (selectedPieces.isEmpty()) { // added HS
	    stopAction.addActionListener(
					 new ActionListener() {
					     public void actionPerformed(ActionEvent e) {
						 abortExecutingAction(selectedPiece);
					     }
					 });
		
	    }
	    // Added HS ////////////////////////////////////////////////////////////////
	    else {
	    stopAction.addActionListener(
					 new ActionListener() {
					     public void actionPerformed(ActionEvent e) {
						 abortExecutingAction(selectedPieces);
					     }
					 });
		
	    }
	    ///////////////////////////////////////////////////////////////////////////////
	    if ((selectedPiece != null)
		&& (selectedPiece.getExecutingAction() == null)) {
		    stopAction.setEnabled(false);
	    }
	    menu.add(stopAction);
	    menu.addSeparator();
	    Enumeration e = actionMenuItems.elements();
	    while (e.hasMoreElements()) {
		menu.add((JMenuItem) e.nextElement());
	    }
	}
	
        return menu;
    }
    /////////////////////////////////////////////////////////////////////////////

    /**
     *  Description of the Method
     */
    private void quit() {
        disconnectFromGame();
        board = null;
        System.exit(0);
    }


    /**
     *  Description of the Method
     */
    private void disconnectFromGame() {
        // abort all actions of all pieces
	Iterator i;
	synchronized (pieces) {
	    i = pieces.values().iterator();
	}
        while (i.hasNext()) {
            Piece p = (Piece) i.next();
            abortExecutingAction(p, true);
        }

        // disconnect and refresh layout etc.
        server.disconnect();
        unselectPiece();
        pieceInfoPanel.setPiece(null);
	synchronized (pieces) {
	    pieces.clear();
	}
        actionExecutingPieces = new Hashtable();
        actionIds = new Hashtable();
        board.clearMap();
        board.repaintBoard();
        setTitle(gameTitle);
    }


    /**
     *  Description of the Method
     *
     * Modified, HS
     */
    private void connectToGame(String serverName, int serverPort,
			       String roleName, boolean showDialog) {
        // connect to server
        if (lastHost == null) {
	    //            lastHost = new HostPort("somecomputer.example.com", 7000);
	    String server = serverName==null? defaultServerName: serverName;
	    int port = serverPort<0? defaultServerPort: serverPort;
            lastHost = new HostPort(server, port);
        }
	HostPort host;
	if (showDialog || serverName==null || serverPort<0){
	    host = ConnectionDialog.showDialog(this, lastHost.getHost(), lastHost.getPort());
	    if (host == null) {
		return;
	    }
	    lastHost = new HostPort(host);
	}
	else {
	    host = new HostPort(lastHost);
	}
        disconnectFromGame();
        boolean couldConnect = server.connect(host.getHost(),
					      host.getPort());
        if (!couldConnect) {
            gecco.client.dialogs.Error.display("Could not connect to that server.",
					       this);
            return;
        }

        // select role
	String role;
        Set roles = server.getAvailableRoles();
	boolean roleNameOK = roles.contains(roleName);
	if (showDialog || roleName==null || !roleNameOK){
	    role = RoleSelectionDialog.showDialog(this, roles);
	    if (role == null) {
		server.disconnect();
		return;
	    }
	}
	else {
	    role = roleName;
	}

        // join as role
        boolean roleOK = server.joinAsRole(role);
        if (!roleOK) {
            gecco.client.dialogs.Error.display("Could not use the role " + role
					       + ". Please try again.", this);
            server.disconnect();
            return;
        }

        // load map
        map = new Map();
        Thread t =
            new Thread() {
                public void run() {
                    map.loadMap(server.getMap());
                    LoadingInfo.hide();
                }
            };
        t.start();
        LoadingInfo.display("Loading map, please wait...", this);

        // launch game
        setTitle(gameTitle + " - " + role);
        board.loadMap(map);
        server.startGame();
        board.repaintBoard();
    }

    private void connectToGame(){
	connectToGame(null, -1, null, true);
    }

    private void connectToGame(String serverName, int serverPort, String roleName){
	connectToGame(serverName, serverPort, roleName, false);
    }
    
    /**
     *  Main method for program start.
     *
     * @param  argv  The command line arguments
     */
    public static void main(String[] args) {

        // Make sure we have java >= 1.4.1.
        VersionNumber requiredVersion = new VersionNumber(1, 4, 1);
        JavaVersion javaVersion = new JavaVersion();
        if (javaVersion.lessThan(requiredVersion)) {
            System.out.println("Running Java version " + javaVersion + ".");
            System.out.println("Java version " + requiredVersion + " or later required. Exiting.");
            System.exit(1);
        }

        // Start the game.
        //Game game = new Game("GECCO");
        Game game = new Game("DKE", new CommandLineProperties(args));
//         game.setVisible(true);
//         game.connectToGame();
    }


    /**
     *  A menu item of the action menu.
     *
     * @author     Andreas Enblom
     * @created    June 27, 2001
     */
    private class ActionMenuItem extends JMenuItem {
        /**
         *  Description of the Field
         */
        Action action;


        /**
         *  Constructor for the ActionMenuItem object
         *
         * @param  a  Description of Parameter
         */
        public ActionMenuItem(Action a) {
            super(a.getName());
            action = a;
        }


        /**
         *  Gets the MyAction attribute of the ActionMenuItem object
         *
         * @return    The MyAction value
         */
        public Action getMyAction() {
            return (Action) action.clone();
        }
    }
}
