/**
 * Represents a unit in the wargame created on Swedish National Defense Collage.
 * The class is based on a class from the original <a href="http://www.nada.kth.se/theory/gecco">
 * GECCO-project</a>
 *
 * @author <a href="mailto:Hans.Sandstrom@fhs.mil.se">Hans Sandstrm</a>
 * @version 1.0
 */

package gecco.game.dke;
import gecco.server.core.*;
import java.util.Iterator;
import java.util.Vector;
import java.util.Enumeration;
import gecco.server.visibility.Point;
import gecco.server.replaymanager.*;
import java.io.ObjectOutputStream;

public abstract class WarUnit extends Unit implements ReplayReady {
 
    private int groundState; // updated in setPosition(double, double)
    private double currentSpeed = 10; // m/s
    private double maxSpeed = 10; // m/s, default value, should be overrided in subclasses
    private Vector zoneOfControll = null; // a Vector containig Points representing current ZC
    private int force; // Blue or red?
    private int underIndirectFireFactor = 0;
    private int disturbedDelayActionHandle = -1;

    private Vector brothersInArms = new Vector();
    private WarUnit supports = null;
    private WarUnit inCombatWhith = null;

    private boolean underAttack = false;
    private boolean attacking = false;
    private boolean firingIndirect = false;
    private boolean underIndirectFire = false;
    private boolean disturbed = false;
    private boolean marching = false;
    private boolean advancing = false;
    private boolean reloading = false;
    public boolean underAttackInitiated = false;
    public boolean attackInitiated = false;
    public boolean regainingStrength = false;

    private int initialFuelLevel = -1; // set in init
    private int initialAmmunitionLevel = -1; // set in init
    private double fuelLevel = -1; // set in init
    
    private int initialHealth = -1; // set in init
    private int lowestHealth = -1; // set in init


    public WarUnit() {
	super();
	force = Def.UNDEFINED_FORCE;
    }

    public void init(){
	initialFuelLevel = getFuelLevelProperty();
	fuelLevel = initialFuelLevel;
	initialAmmunitionLevel = getAmmunitionLevel();
	initialHealth = getHealth();
	lowestHealth = initialHealth;
    }

    public void markAsDestroyed(){
	//ReferenceHolder.getReferee().unitDestroyed(this);
	sendUnitMessage(Def.getUnitDestroyedWarningString());
	super.markAsDestroyed();
    }
    
    public int getGroundState(){
	return groundState;
    }

    /**
     * Returns the time that shall pass before next update of movement. The method 
     * uses <code>Def.STEP_LENGTH, Def.MAP_SCALE</code>, the property "Rrlighet"
     * and <code>groundState</code>
     *
     * @return seconds to next update
     */
    double getTimeToRequeueForMovement() {
	return getTimeToRequeueForMovement(Def.STEP_LENGTH);
    }
    
    /**
     * Returns the time that shall pass before next update of movement. The method 
     * uses <code>Def.STEP_LENGTH, Def.MAP_SCALE</code>, the property "Rrlighet"
     * and <code>groundState</code>
     *
     * @return seconds to next update
     */
    double getTimeToRequeueForMovement(double stepLength) {
	String speedLevel = getStringProperty(Def.getMovementPropertyString());
	double disturbedFactor = isDisturbed()? Def.getDisturbedPerformanceFactor(): 1;
	double terrainDenominator = Def.getTerrainSpeedDenominator(groundState);
	double speedFactor = Def.speedFactorConverter(getSpeedFactor());
	double fuelFactor = getFuelLevelSpeedFactor();
	double currentSpeed = Def.getSpeed(speedLevel)*disturbedFactor/terrainDenominator*speedFactor*fuelFactor;
	double secondsToUpdate = stepLength * Def.MAP_SCALE / currentSpeed;
	return secondsToUpdate;
    }
    
    public void setPosition(double x, double y){
	setPosition(x, y, false);
    }

    public void setPosition(double x, double y, boolean autoAttack) {
	super.setPosition(x,y);
	groundState = ReferenceHolder.getAutomatonMap().getAutomaton((int)x,(int)y).getState();
	// Test for and perform automatic attack
	if (autoAttack && (getAttackFactor() > 0) && !isDestroyed() 
	    && !attackInitiated && !underAttackInitiated) {
	    Iterator it = ReferenceHolder.getUnitManager().getAllUnits().iterator();
	    boolean go = true;
	    while (it.hasNext() && go) {
		Object o = it.next();
		if (o instanceof WarUnit ) {
		    WarUnit wu = (WarUnit) o;
		    if (!wu.isDestroyed() && (wu.getForce() != this.getForce())) {
			// Check if we're inside the attack radius!
			double dx = wu.getX() - getX();
			double dy = wu.getY() - getY();
			double dist = Math.sqrt(dx*dx + dy*dy);
			if (dist <  Def.getAttackRadius()) {
			    // Debug info
			    //System.out.println("WarUnit: " + getName() + " brjar automatisk attack");
			    
			    go = false;
			    new AutoAttackAction(this.getHandle(), wu.getHandle());
			}
		    }
		}
	    }
	}
	// Don't record replay data when replaying
	if (!ReferenceHolder.inReplayMode()) {
	    Object[] parameters = {new Double(x), new Double(y)};
	    saveReplayData(SET_POSITION, parameters);
	}
    }

    public void setAttacking(boolean b){
	attacking = b;
	// Don't record replay data when replaying
	if (!ReferenceHolder.inReplayMode()) {
	    Object[] parameters = {new Boolean(b)};
	    saveReplayData(SET_ATTACKING, parameters);
	}
    }
    public void setUnderAttack(boolean b){
	underAttack = b;
	// Don't record replay data when replaying
	if (!ReferenceHolder.inReplayMode()) {
	    Object[] parameters = {new Boolean(b)};
	    saveReplayData(SET_UNDER_ATTACK, parameters);
	}
    }
    public boolean getAttacking(){
	return attacking;
    }
    public boolean getUnderAttack(){
	return underAttack;
    }
    public void setSupports(WarUnit wU){
	supports = wU;
    }
    public WarUnit getSupports(){
	return supports;
    }

    /**
     * Starts an attack on target.
     * if target is allready in combat, that combat
     * is joined.
     * Returns true if there wasn't another combat going on.
     *
     * @param target a <code>WarUnit</code> value
     * @return a <code>boolean</code> value
     */
    public boolean attack(WarUnit target) {
	WarUnit wU = target.initUnderAttack(this);
	if (wU == null) {
	    setInCombatWhith(target);
	    setAttacking(true);
	    updateStatus();
	    return true;
	}
	else {
	    wU.joinForCombat(this);
	    setSupports(wU);
	    setInCombatWhith(wU.getInCombatWhith());
	    setAttacking(wU.getAttacking());
	    setUnderAttack(wU.getUnderAttack());
	    updateStatus();
	    return false;
	}
    }

    public WarUnit initUnderAttack(WarUnit attacker){
	if (isInCombat()) {
	    return getInCombatWhith();
	}
	else {
	    setInCombatWhith(attacker);
	    setUnderAttack(true);
	    sendUnitMessage(Def.getCombatWarningString());
	    new UpdateStatusAction(this.getHandle());
	    return null;
	}
    }

    public void joinForCombat(WarUnit brotherInArms){
	brothersInArms.add(brotherInArms);
    }

    public void combatEnded(){
	// Denna funktion SKALL endast anropas p den enhet som leder  en attack
	
	// Debug info
	if (!getAttacking() || !(getInCombatWhith().getUnderAttack())) {
	    System.out.println("WarUnit.combatEnded() Ngot har gtt snett, getAttacking() = " +
			       getAttacking() + " getInCombatWhith().getUnderAttack() = " +
			       getInCombatWhith().getUnderAttack());
	    System.exit(1);
	}
	int attackValue = getAttackValue();
	int defenceValue = getInCombatWhith().getDefenceValue();
	int[] result = Def.getResutltOfCombat(attackValue, defenceValue);

	double enemyX = getInCombatWhith().getX();
	double enemyY = getInCombatWhith().getY();
	////////////////////////////////////////////////////////////////////////
	//  Hmtat frn actionProcessor, i stora drag                         //
	////////////////////////////////////////////////////////////////////////
	WarUnit wU = getInCombatWhith();
	double oldx = wU.getX();
	double oldy = wU.getY();
	wU.resetUpdatedActions();
	wU.resetUpdatedProperties();
	
	wU.combatResult(result[1], result[0], getX(), getY());
	
	// Check if things has changed...
	if (wU.isDestroyed()) {
	    ReferenceHolder.getUnitManager().deleteUnit(wU, oldx, oldy);
	    ReferenceHolder.getVisibilityFilter().deleteUnit(wU, oldx, oldy);
	} else {
	    ReferenceHolder.getActionProcessor().publicCheckChangedPosition(oldx, oldy, wU);
	    ReferenceHolder.getActionProcessor().publicCheckChangedProperties(wU);
	    ReferenceHolder.getActionProcessor().publicCheckActionBlocklist(wU);
	}
	///////////////////////////////////////////////////////////////////////

	// If the defender retreats or is eliminated the main attacker must advance
	if ((result[1] & Def.RETREAT) == Def.RETREAT ||
	    getInCombatWhith().isDestroyed()) {
	    result[0] = result[0] | Def.ADVANCE;
	}
	combatResult(result[0], result[1], enemyX, enemyY);
    }

    //     public void combatEnded(int result, double enemyX, double enemyY){
    // 	combatResult(result, enemyX, enemyY);
    //     }

    public void combatResult(int result, int enemyResult, double enemyX, double enemyY){
	int damage = result & Def.RESULT_MASK;
	if (!brothersInArms.isEmpty()) {
	    int damageShare = damage / (brothersInArms.size()+ 1); // Heltalsdivision
	    // Keep eliminated and retreat part and add share of damage
	    // Only the main attacker should advance, therfore remove advance part
	    int subResult = (result & ~Def.RESULT_MASK & ~Def.ADVANCE) + damageShare; 
	    Enumeration e = brothersInArms.elements();
	    while (e.hasMoreElements()) {
		WarUnit wU = (WarUnit)e.nextElement();
		////////////////////////////////////////////////////////////////////////
		//  Hmtat frn actionProcessor, i stora drag                         //
		////////////////////////////////////////////////////////////////////////
                double oldx = wU.getX();
                double oldy = wU.getY();
                wU.resetUpdatedActions();
                wU.resetUpdatedProperties();
		
		wU.combatResult(subResult, enemyResult, enemyX, enemyY);

                // Check if things has changed...
                if (wU.isDestroyed()) {
		    ReferenceHolder.getUnitManager().deleteUnit(wU, oldx, oldy);
		    ReferenceHolder.getVisibilityFilter().deleteUnit(wU, oldx, oldy);
                } else {
                    ReferenceHolder.getActionProcessor().publicCheckChangedPosition(oldx, oldy, wU);
                    ReferenceHolder.getActionProcessor().publicCheckChangedProperties(wU);
                    ReferenceHolder.getActionProcessor().publicCheckActionBlocklist(wU);
                }
		///////////////////////////////////////////////////////////////////////

	    }
	    // main attacker/defender has to take the damage that can't be spread evenly
	    damage = (damage % (brothersInArms.size()+ 1)) + damageShare;
	    brothersInArms.clear();
	}
	// Effekter av strid
	boolean eliminated = (result & Def.ELIMINATED) == Def.ELIMINATED;
	boolean retreat = (result & Def.RETREAT) == Def.RETREAT;
	boolean advance = (result & Def.ADVANCE) == Def.ADVANCE;
	if (eliminated) {
	    decreaseHealth(getHealth());
	}
	else {
	    if (retreat && !isDestroyed()) {
		retreatAfterCombat(enemyX, enemyY);
	    }
	    if (advance && !isDestroyed()) {
		// advanceAfterCombat(enemyX, enemyY);
	    }
	    decreaseHealth(damage);
	}

	// terstll
	setInCombatWhith(null);
	setSupports(null);
	setAttacking(false);
	setUnderAttack(false);

	// Resursfrbrukning
	decreaseFuelLevel(Def.getFuelConsumptionCombat());
	consumeAmmunition();

	// Rapport
	sendCombatEndedMessage(damage, enemyResult & Def.RESULT_MASK);

	updateStatus();
    }

    public void sendCombatEndedMessage(int ownDamage, int enemyDamage){
	String s = Def.getCombatEndedString() + "\n" +
	    Def.getEnemyLossesString() + enemyDamage + "\n" +
	    Def.getOwnLossesString() + ownDamage + "\n";
	sendUnitMessage(s);
    }

    public void retreatAfterCombat(double enemyX, double enemyY){
	// Debug info
	//System.out.println("WarUnit.retreatAfterCombat()");
	double curX = getX();
	double curY = getY();
	double baseX = Def.getRetreatBase(getForce())[0];
	double baseY = Def.getRetreatBase(getForce())[1];
	double dxBase = baseX - curX;
	double dyBase = baseY - curY;
	double dxEnemy = enemyX - curX;
	double dyEnemy = enemyY - curY;
	double angleToBase = Math.atan2(dxBase, dyBase);
	double angleToEnemy = Math.atan2(dxEnemy, dyEnemy);
	double retreatAngle;
	if (angleToBase < 0) {
	    angleToBase += 2*Math.PI;
	}
	if (angleToEnemy < 0) {
	    angleToEnemy += 2*Math.PI;
	}
	
	double angleDistance = angleToBase - angleToEnemy;
	if ((angleDistance > Math.PI/2 && angleDistance < Math.PI*3/2) ||
	    (angleDistance < -Math.PI/2 && angleDistance > -Math.PI*3/2)) {
	    // Debug info
	    //System.out.println("Vinkel OK, " + Math.toDegrees(angleDistance));
	    retreatAngle = angleToBase;
	}
	else {
	    if (angleDistance > 0) {
		retreatAngle = angleToBase + (Math.PI/2 - angleDistance);
	    }
	    else {
		retreatAngle = angleToBase - (Math.PI/2 + angleDistance);
	    } 
	}
	double retreatDistance = 2 * Def.getAttackRadius();
	while (retreatDistance > 0) {
	    double x = getX() + retreatDistance*Math.sin(retreatAngle);
	    double y = getY() + retreatDistance*Math.cos(retreatAngle);
	    int intX = (int) x;
	    int intY = (int) y;
	    if (ReferenceHolder.getAutomatonMap().areValidCoordinates(intX, intY) && !isOutsideMap(x,y)) {
		int automatonState = ReferenceHolder.getAutomatonMap().getAutomaton(intX, intY).getState();
		if (automatonState != Def.SEA) {
		    // OK to retreat
		    break;
		}
	    }
	    // Seams to retreating to the sea or to invalid map coordinates
	    retreatDistance --;
	}
	if (retreatDistance < 0){
	    retreatDistance = 0;
	}
	
	// There are something funny whith the coordinat system
	performRetreatTo(getX() + retreatDistance*Math.sin(retreatAngle),
			 getY() + retreatDistance*Math.cos(retreatAngle));
	return;
    }

    public void performRetreatTo(double x, double y){
	super.setPosition(x,y);
	// Don't record replay data when replaying
	if (!ReferenceHolder.inReplayMode()) {
	    Object[] parameters = {new Double(x), new Double(y)};
	    saveReplayData(PERFORM_RETREAT_TO, parameters);
	}
    }

    public void advanceAfterCombat(double enemyX, double enemyY){
	setPosition(enemyX, enemyY, false);
	return;
    }

    public int getAttackValue(){
	int value = 0;
	if (!brothersInArms.isEmpty()) {
	    Enumeration e = brothersInArms.elements();
	    while (e.hasMoreElements()) {
		WarUnit wU = (WarUnit)e.nextElement();
		value += wU.getAttackValue();
	    }
	}
	double disturbedFactor = isDisturbed()? Def.getDisturbedPerformanceFactor(): 1;
	double ammunitionLevelFactor = getAmmunitionLevelAttackFactor();
	value += (int)(getAttackFactor() * disturbedFactor * ammunitionLevelFactor);
	return value;
    }

    public int getDefenceValue(){
	int value = 0;
	if (!brothersInArms.isEmpty()) {
	    Enumeration e = brothersInArms.elements();
	    while (e.hasMoreElements()) {
		WarUnit wU = (WarUnit)e.nextElement();
		value += wU.getDefenceValue();
	    }
	}
	double disturbedFactor = isDisturbed()? Def.getDisturbedPerformanceFactor(): 1;
	double terrainFactor = Def.getTerrainAttackProtectionFactor(groundState);
	double ammunitionLevelFactor = getAmmunitionLevelDefenceFactor();
	value += (int)(getDefenceFactor() * terrainFactor * disturbedFactor * ammunitionLevelFactor);
	return value;
    }
	
    public void resetCombatants(){
	if (!brothersInArms.isEmpty()) {
	    Enumeration e = brothersInArms.elements();
	    while (e.hasMoreElements()) {
		WarUnit wU = (WarUnit)e.nextElement();
		wU.resetCombatants();
	    }
	    brothersInArms.clear();
	}
	setInCombatWhith(null);
	setSupports(null);
	setAttacking(false);
	setUnderAttack(false);
	updateStatus();
    }

    public boolean isInCombat(){
	return (attacking || underAttack);
    }

    public WarUnit getInCombatWhith(){
	return inCombatWhith;
    }

    public void setInCombatWhith(WarUnit u){
	inCombatWhith = u;
    }

    public int getSpeedFactor(){
	return getIntProperty(Def.getSpeedFactorString());
    }

    public void setSpeedFactor(int n){
	setProperty(Def.getSpeedFactorString(), n);
	// Don't record replay data when replaying
	if (!ReferenceHolder.inReplayMode()) {
	    Object[] parameters = {new Integer(n)};
	    saveReplayData(SET_SPEED_FACTOR, parameters);
	}
    }

    public int getAttackFactor(){
	return (getIntProperty(Def.getAttackPropertyString()));
    }

    public void setAttackFactor(int n){
	setProperty(Def.getAttackPropertyString(), n);
    }   

    public int decreaseAttackFactor(int n){
	int oldAttackFactor = getAttackFactor();
	if (oldAttackFactor == 0) { // Enheter som inte har en attackfaktor skall inte kunna ka den
	    return oldAttackFactor;
	}
	int newAttackFactor = oldAttackFactor - n;
	if (newAttackFactor < 1) {
	    newAttackFactor = 0;
	}
	setAttackFactor(newAttackFactor);
	return newAttackFactor;
    }

    public int getDefenceFactor(){
	return getIntProperty(Def.getDefencePropertyString());
    }

    public void setDefenceFactor(int n){
	setProperty(Def.getDefencePropertyString(), n);
    }

    public int decreaseDefenceFactor(int n){
	int oldDefenceFactor = getDefenceFactor();
	int newDefenceFactor = oldDefenceFactor - n;
	if (newDefenceFactor < 1) {
	    newDefenceFactor = 1;
	}
	setDefenceFactor(newDefenceFactor);
	return newDefenceFactor;
    }

    public int getIndirectFireFactor(){
	return getIntProperty(Def.getIndirectFirePropertyString());
    }

    public void setIndirectFireFactor(int n){
	setProperty(Def.getIndirectFirePropertyString(), n);
    }

    public int decreaseIndirectFireFactor(int n){
	int oldIndirectFireFactor = getIndirectFireFactor();
	if (oldIndirectFireFactor == 0) { // Enheter som inte har indirekt eld skall inte kunna ka den
	    return oldIndirectFireFactor;
	}
	int newIndirectFireFactor = oldIndirectFireFactor - n;
	if (newIndirectFireFactor < 1) {
	    newIndirectFireFactor = 0;
	}
	setIndirectFireFactor(newIndirectFireFactor);
	return newIndirectFireFactor;
    }

    public void releaseFiringIndirect(){
	setFiringIndirect(false);
	updateStatus();
    }

    public int getHealth(){
	return (getIntProperty(Def.getHealthPropertyString()));
    }

    public void setHealth(int n){
	setProperty(Def.getHealthPropertyString(), n);
    }

    public int decreaseHealth(int n){
	// Debug info
	// System.out.println("WarUnit.decreaseHealth("+n+")");
	int oldHealth = getHealth();
	int newHealth = oldHealth - n;
	if (newHealth < 1) {
	    markAsDestroyed();
	    newHealth = 0;
	}
	lowestHealth = Math.min(lowestHealth, newHealth);
	setHealth(newHealth);
	decreaseAttackFactor(n);
	decreaseDefenceFactor(n);
	decreaseIndirectFireFactor(n);

	// Don't record replay data when replaying
	if (!ReferenceHolder.inReplayMode()) {
	    Object[] parameters = {new Integer(n)};
	    saveReplayData(DECREASE_HEALTH, parameters);
	}

	return newHealth;
    }
    
    public void setFuelLevelProperty(int n){
	setProperty(Def.getFuelPropertyString(), n);
    }
    public int getFuelLevelProperty(){
	return getIntProperty(Def.getFuelPropertyString());
    }
    public double getFuelLevel(){
	return fuelLevel;
    }
    public void setFuelLevel(double x){
	fuelLevel = x;
    }
    public void decreaseFuelLevel(double x){
	// Debug info
	// System.out.println("WarUnit.decreaseFuelLevel("+x+")");
	int oldIntLevel = getIntProperty(Def.getFuelPropertyString());
	double oldFuelLevel = getFuelLevel();
	double newFuelLevel = oldFuelLevel - x;
	if (newFuelLevel > initialFuelLevel) {
	    newFuelLevel = initialFuelLevel;
	}
	if (newFuelLevel < 0) {
	    newFuelLevel = 0;
	}
	setFuelLevel(newFuelLevel);
	int newIntLevel = (int) newFuelLevel;
	if (!(newIntLevel == oldIntLevel)) {
	    setProperty(Def.getFuelPropertyString(), newIntLevel);
	}
	// Don't record replay data when replaying
	if (!ReferenceHolder.inReplayMode()) {
	    Object[] parameters = {new Double(x)};
	    saveReplayData(DECREASE_FUEL_LEVEL, parameters);
	}
    }
    public double getFuelLevelSpeedFactor(){
	double level = getFuelLevel();
	if (level < Def.getCriticalFuelLevel()) {
	    return Def.getCriticalFuelLevelSpeedFactor();
	}
	else if (level < Def.getLowFuelLevel()) {
	    return Def.getLowFuelLevelSpeedFactor();
	}
	else {
	    return 1;   
	}
    }
    public void fuelConsumtion(double distance){
	decreaseFuelLevel(distance*0.1);
    }
    public void fuelConsumtion(){
	fuelConsumtion(Def.getStepLength());
    }

    public void setAmmunitionLevel(int n){
	setProperty(Def.getAmmunitionPropertyString(), n);
    }
    public int getAmmunitionLevel(){
	return getIntProperty(Def.getAmmunitionPropertyString());
    }
    public void decreaseAmmunitionLevel(int n){
	// Debug info
	// System.out.println("WarUnit.decreaseAmmunitionLevel("+n+")");
	int newLevel = getAmmunitionLevel() - n;
	if (newLevel < 0) newLevel = 0;
	if (newLevel > initialAmmunitionLevel) newLevel = initialAmmunitionLevel;
	setAmmunitionLevel(newLevel);
	// Don't record replay data when replaying
	if (!ReferenceHolder.inReplayMode()) {
	    Object[] parameters = {new Integer(n)};
	    saveReplayData(DECREASE_AMMUNITION_LEVEL, parameters);
	}
    }
    public double getAmmunitionLevelAttackFactor(){
	if (getAmmunitionLevel() < Def.getLowAmmunitionLevel()) {
	    return Def.getLowAmmunitionLevelAttackFactor();
	}else {
	    return 1;
	}
    }
    public double getAmmunitionLevelDefenceFactor(){
	if (getAmmunitionLevel() < Def.getCriticalAmmunitionLevel()) {
	    return Def.getCriticalAmmunitionLevelDefenceFactor();
	}else {
	    return 1;
	}
    }
    public void consumeAmmunition(){
	int min = Def.getAmmunitonConsumptionMin();
	int spread = Def.getAmmunitonConsumptionSpread();
	int consumption = min + (int) (Math.random()*(spread+1));
	decreaseAmmunitionLevel(consumption);
    }

    public void regainStrength(){
	removeProperty(Def.getRegainStrengthPropertyString());
	decreaseAmmunitionLevel(-Def.getRegainAmmunitionValue());
	decreaseFuelLevel(-Def.getRegainFuelValue());
	int newHealth = (int) (lowestHealth * Def.getRegainHealthFactor());
	newHealth = Math.min(newHealth, initialHealth);
	decreaseHealth(-(newHealth - getHealth()));
    }

    public void setFiringIndirect(boolean b){
	firingIndirect = b;
	// Don't record replay data when replaying
	if (!ReferenceHolder.inReplayMode()) {
	    Object[] parameters = {new Boolean(b)};
	    saveReplayData(SET_FIRING_INDIRECT, parameters);
	}
    }
    public void setUnderIndirectFire(boolean b){
	underIndirectFire = b;
	// Don't record replay data when replaying
	if (!ReferenceHolder.inReplayMode()) {
	    Object[] parameters = {new Boolean(b)};
	    saveReplayData(SET_UNDER_INDIRECT_FIRE, parameters);
	}
    }
    public void setDisturbed(boolean b){
	disturbed = b;
	// Don't record replay data when replaying
	if (!ReferenceHolder.inReplayMode()) {
	    Object[] parameters = {new Boolean(b)};
	    saveReplayData(SET_DISTURNED, parameters);
	}
    }
    public void setMarching(boolean b){
	marching = b;
	// Don't record replay data when replaying
	if (!ReferenceHolder.inReplayMode()) {
	    Object[] parameters = {new Boolean(b)};
	    saveReplayData(SET_MARCHING, parameters);
	}
    }
    public void setAdvancing(boolean b){
	advancing = b;
	// Don't record replay data when replaying
	if (!ReferenceHolder.inReplayMode()) {
	    Object[] parameters = {new Boolean(b)};
	    saveReplayData(SET_ADVANCING, parameters);
	}
    }
    public void setReloading(boolean b){
	reloading = b;
	// Don't record replay data when replaying
	if (!ReferenceHolder.inReplayMode()) {
	    Object[] parameters = {new Boolean(b)};
	    saveReplayData(SET_RELOADING, parameters);
	}
    }
    public void setRegainingStrength(boolean b){
	regainingStrength = b;
	updateStatus();
	// Don't record replay data when replaying
	if (!ReferenceHolder.inReplayMode()) {
	    Object[] parameters = {new Boolean(b)};
	    saveReplayData(SET_REGAINING_STRENGTH, parameters);
	}
    }

    public boolean isDisturbed(){
	return disturbed;
    }
    public boolean isMarching(){
	return marching;
    }
    public boolean isAdvancing(){
	return advancing;
    }
    public boolean isUnderIndirectFire(){
	return underIndirectFire;
    }
    public boolean isFiringIndirect(){
	return firingIndirect;
    }
    public boolean isAttacking(){
	return attacking;
    }
    public boolean isUnderAttack(){
	return underAttack;
    }
    public boolean isReloading(){
	return reloading;
    }
    public boolean isRegainingStrength(){
	return regainingStrength;
    }

    public void updateStatus(){
	if (!underAttack && !underIndirectFire && !attacking && !firingIndirect &&
	    !disturbed && !marching && !advancing && !reloading && !regainingStrength) {
	    setProperty(Def.getStatusPropertyString(), " ");
	}
	else {
	    String statusString = (disturbed ? "Disturbed ": "") + (attacking ? "Attacking ": "") +
		(reloading ? "Reloading ": "") + (regainingStrength ? "Recovering ": "") +
		(underAttack ? "Under attack ": "") + (firingIndirect ? "Firing indirect ": "") +
		(underIndirectFire ? "Under fire ": "") + (marching ? "Moving ": "") +
		(advancing ? "Advancing ": "");
	    setProperty(Def.getStatusPropertyString(), statusString);
	}
	// Don't record replay data when replaying
	if (!ReferenceHolder.inReplayMode()) {
	    Object[] parameters = {null};
	    saveReplayData(UPDATE_STATUS, parameters);
	}
    }

    public void setReloadProperty(String s){
	setProperty(Def.getReloadPropertyString(), s);
	// Don't record replay data when replaying
	if (!ReferenceHolder.inReplayMode()) {
	    Object[] parameters = {s};
	    saveReplayData(SET_RELOAD_PROPERTY, parameters);
	}
    }	

    public void setRegainStrengthProperty(String s){
	setProperty(Def.getRegainStrengthPropertyString(), s);
	// Don't record replay data when replaying
	if (!ReferenceHolder.inReplayMode()) {
	    Object[] parameters = {s};
	    saveReplayData(SET_REGAIN_STRENGTH_PROPERTY, parameters);
	}
    }	

    public int getForce(){
	return force;
    }

    public void setForce(int force){
	this.force = force;
    }

    public double getReloadExtendTime(){
	return Def.getReloadExtendTime(this);
    }

    public double getIndirectFireExtendTime(){
	return Def.getIndirectFireExtendTime(this);
    }

    public int getIndirectFireDispersion(){
	return Def.getIndirectFireDispersion(this);
    }

    public void underIndirectFire(int factor){
	boolean startOfFire = false;
	if (!underIndirectFire) {
	    startOfFire = true;
	    sendUnitMessage(Def.getIndirectFireWarningString());
	    if (!isAttacking() && !isUnderAttack()) {
		//abortAllActions();		
	    }
	}
	underIndirectFireFactor += factor;
	setUnderIndirectFire(true);
	updateStatus();
	if (startOfFire) {
	    new UnderIndirectFireDelayAction(this.getHandle(), getTimeToShowIndirectFire());
	}
    }

    public double getTimeToShowIndirectFire(){
	return Def.getTimeToShowIndirectFire();
    }

    /**
     * Calculates the effect of indirect fire
     *
     */
    public void underIndirectFireCompleted(){
	int protectionFactor = Def.getProtection(getStringProperty(Def.getProtectionPropertyString()));
	double terrainFactor = Def.getTerrainIndirectFireProtectionFactor(groundState);
	int indirectFireProtection = (int)(protectionFactor * terrainFactor);
	int result = Def.getResultOfArtilleryFire(underIndirectFireFactor, indirectFireProtection);
	boolean disturbEffect = (result & Def.DISTURBED) == Def.DISTURBED;
	int damageResult = result & Def.RESULT_MASK;
	decreaseHealth(damageResult);
	if (!isDestroyed() && disturbEffect){
	    setDisturbed(disturbEffect);
	    int actionHandle = gecco.server.core.Action.getNewActionHandle();
	    disturbedDelayActionHandle = actionHandle;
	    DisturbedDelayAction dDA = new DisturbedDelayAction();
	    dDA.initiate(actionHandle, this.getHandle(), this.getHandle(), getDisturbedTime());
	    ReferenceHolder.getActionProcessor().initiateAction(Def.getSystemRoleName(), dDA);
	}
	underIndirectFireFactor = 0;
	setUnderIndirectFire(false);
	updateStatus();
    }

    public void resetDisturbed(int actionHandle){
	if (actionHandle == disturbedDelayActionHandle){
	    setDisturbed(false);
	    updateStatus();
	}
    }

    public double getDisturbedTime(){
	return Def.getDisturbedTime();
    }

    public double getCombatExtendTime(){
	return Def.getCombatExtendTime();
    }

    public int getRegainStrengthTime(){
	return Def.getRegainStrengthTime();
    }

    public void sendUnitMessage(String s){
	super.sendUnitMessage(s);
	// Don't record replay data when replaying
	if (!ReferenceHolder.inReplayMode()) {
	    Object[] parameters = {s};
	    saveReplayData(SEND_UNIT_MESSAGE, parameters);
	}
    }

    public boolean cantMoveTo(int automatonState){
	if (automatonState == Def.SEA) {
	    return true;
	}
	else {
	    return false;
	}
    }

    public boolean isEnemyForcesWithinRangeOfUnit(Vector excludedUnits){
	Vector enemiesInRange = getEnemyForcesWithinRangeOfUnit();
	if (excludedUnits == null) {
	    excludedUnits = new Vector();
	}
	enemiesInRange.removeAll(excludedUnits);
	return !(enemiesInRange.isEmpty());
    
    }

    public boolean isEnemyForcesWithinRangeOfUnit(){
	return isEnemyForcesWithinRangeOfUnit(null);
    }

    /**
     * Returns a Vector containing the WarUnit object of the enemy forces
     * whitin the range of the WarUnit
     *
     * @return a <code>Vector</code> value
     */
    public Vector getEnemyForcesWithinRangeOfUnit(){
	int unitForce = getForce();
	double rangeSquare = getUnitType().getRangeSquare();
	double unitX = getX();
	double unitY = getY();
	Iterator it = ReferenceHolder.getUnitManager().getAllUnits().iterator();
	double x, y, dx, dy;
	WarUnit wU;
	Unit u;
	Vector enemyForcesWhitinRange = new Vector();

	while (it.hasNext()) {
	    u = (Unit)it.next();
	    if (u instanceof WarUnit) {
		wU = (WarUnit)u;
		if (!(unitForce == wU.getForce())){
		    x = wU.getX();
		    y = wU.getY();
		    dx = unitX - x;
		    dy = unitY - y;
		    if (rangeSquare >= (dx*dx + dy*dy)){
			enemyForcesWhitinRange.add(wU);
		    }
		}
	    }
	}
	return enemyForcesWhitinRange;
    }

    /**
     * Retruns a Vector containing Points representing the 
     * Automatons that the Unit can see
     *
     * @return a <code>Vector</code> value
     */
    Vector getPointsVisibleToUnit(){
	Vector pointV = new Vector();
	int range = getUnitType().getRange();
	int curX = getIntX();
	int curY = getIntY();
	for (int y = -range; y < range+1; y++) {
	    for (int x = -range; x < range+1; x++) {
		if (ReferenceHolder.getAutomatonMap().areValidCoordinates(x+curX, y+curY)) {
		    if (canSeeAutomatonAtPosition(x+curX, y+curY)) {
			pointV.add(new Point(x+curX, y+curY));
		    }
		}
	    }
	}
	return pointV;
    }

	


    public void updateZoneOfControll(){
	// There are no need to remove any zc, just add all of them
	if (zoneOfControll == null) {
	    zoneOfControll = new Vector(100);
	    Enumeration e = getPointsVisibleToUnit().elements();
	    AutomatonImpl aI;
	    Point p;
	    while (e.hasMoreElements()) {
		p = (Point)e.nextElement();
		zoneOfControll.add(p);
		aI =(AutomatonImpl)ReferenceHolder.getAutomatonMap().getAutomaton(p.x, p.y);
		aI.setZC(true, getForce(), getHandle());
	    } 
	    return;
	}
	
	// Check whitch zc that has to be set, then remove the
	// ones that isn't valid any more
	Vector v1 = getPointsVisibleToUnit();
	Vector newZC = (Vector)v1.clone();
	v1.removeAll(zoneOfControll);
	Enumeration e = v1.elements();
	AutomatonImpl aI;
	Point p;
	while (e.hasMoreElements()) {
	    p = (Point) e.nextElement();
	    aI =(AutomatonImpl)ReferenceHolder.getAutomatonMap().getAutomaton(p.x, p.y);
	    aI.setZC(true, getForce(), getHandle());
	}
	zoneOfControll.removeAll(newZC);
	e = zoneOfControll.elements();
	while (e.hasMoreElements()) {
	    p = (Point) e.nextElement();
	    aI = (AutomatonImpl)ReferenceHolder.getAutomatonMap().getAutomaton(p.x, p.y);
	    aI.setZC(false, getForce(), getHandle());
	}
	zoneOfControll = newZC;
	return;
    }

    public  boolean isOutsideMap(double x, double y) {
        gecco.server.automaton.AutomatonMap map = ReferenceHolder.getAutomatonMap();
	// Kolla x
        double maxX = (double) map.getWidth();
        if (x < 0.0 || x >= maxX) {
            return true;
	}
	// Kolla y
        double maxY = (double) map.getHeight();
        if (y < 0.0 || y >= maxY) {
            return true;
	}
	// ok
	return false;
    }
   
    /**
     *
     * @return description of instance
     */
    public String toString(){
	return super.toString() +
	    "\nForce         : " + Def.getForceName(getForce()) +
	    "\nHealth        : " + getHealth();
    }

    public void printWriteMe(java.io.PrintWriter out){
	super.printWriteMe(out);
	out.println("Force         : " + Def.getForceName(getForce()));
	out.println("Health        : " + getHealth());
    }


    // implementation of gecco.server.core.ReplayReady interface

    /**
     *
     * @param param1 <description>
     * @param param2 <description>
     *
     * Format description: TimeStamp Type Handle Method Parameters
     * TimeStamp  type: double    description: game time when method was called
     * Type       type: int       description: what type of object is executing the method
     *                                     currently only 1 = Unit is valid
     * Handle     type: int       description: the handle for the type, currently unitHandle
     * Method     type: int       description: which method is called, handled by the implementor
     * Parameters type: Object[]  description: parameters to the method embedded in a Object[]
     *
     *
     * OBSERVE! Since replayLog is a ObjectOutputSteam all objects that should be written must 
     * implement the java.io.Serializable interface
     *
     */
    public void saveReplayData(int method, Object[] parameters){
	synchronized (ReferenceHolder.getReplayLog()){
	    ObjectOutputStream replayLog = ReferenceHolder.getReplayLog();
	    try {
		replayLog.writeDouble(ReferenceHolder.getClock().gameTimeNow());
		replayLog.writeInt(UNIT);
		replayLog.writeInt(getHandle());
		replayLog.writeInt(method);
		replayLog.writeObject(parameters);
		replayLog.flush();
	    } catch (Exception e) {
		e.printStackTrace();
	    } 
	}
    }
	
    private final int SET_POSITION = 1;
    private final int DECREASE_HEALTH = 2;
    private final int SET_FIRING_INDIRECT = 3;
    private final int SET_UNDER_INDIRECT_FIRE = 5;
    private final int SET_DISTURNED = 6;
    private final int SET_MARCHING = 7;
    private final int SET_ADVANCING = 8;
    private final int SET_RELOADING = 9;
    private final int UPDATE_STATUS = 10;
    private final int SET_ATTACKING = 11;
    private final int SET_UNDER_ATTACK = 12;
    private final int SET_SPEED_FACTOR = 13;
    private final int DECREASE_FUEL_LEVEL = 14;
    private final int SET_RELOAD_PROPERTY = 15;
    private final int DECREASE_AMMUNITION_LEVEL = 16;
    private final int SET_REGAINING_STRENGTH = 17;
    private final int SET_REGAIN_STRENGTH_PROPERTY = 18;
    private final int SEND_UNIT_MESSAGE = 19;
    private final int PERFORM_RETREAT_TO = 20;

    /**
     *
     * @param method the method that should be called
     * @param parameters an <code>Object[]</code> containing the parameters for the method
     */
    public void playReplayData(int method, Object[] parameters){
	boolean b;
	int n;
	double d;
	String s;
	double x;
	double y;
 	switch (method) {
	case SET_POSITION :
	    x = ((Double) parameters[0]).doubleValue();
	    y = ((Double) parameters[1]).doubleValue();
	    setPosition(x,y);
	    break;
	case DECREASE_HEALTH :
	    n = ((Integer) parameters[0]).intValue();
	    decreaseHealth(n);
	    break;
	case SET_SPEED_FACTOR :
	    n = ((Integer) parameters[0]).intValue();
	    setSpeedFactor(n);
	    break;
	case SET_FIRING_INDIRECT :
	    b =((Boolean) parameters[0]).booleanValue();   
	    setFiringIndirect(b);
	    break;
	case SET_UNDER_INDIRECT_FIRE :
	    b =((Boolean) parameters[0]).booleanValue();   
	    setUnderIndirectFire(b);
	    break;
	case SET_DISTURNED :
	    b =((Boolean) parameters[0]).booleanValue();   
	    setDisturbed(b);		
	    break;
	case SET_MARCHING :
	    b =((Boolean) parameters[0]).booleanValue();   
	    setMarching(b);
	    break;
	case SET_ADVANCING :
	    b =((Boolean) parameters[0]).booleanValue();   
	    setAdvancing(b);
	    break;
	case SET_RELOADING :
	    b =((Boolean) parameters[0]).booleanValue();   
	    setReloading(b);
	    break;
	case UPDATE_STATUS :
	    updateStatus();
	    break;
	case SET_ATTACKING :
	    b =((Boolean) parameters[0]).booleanValue();   
	    setAttacking(b);
	    break;
	case SET_UNDER_ATTACK :
	    b =((Boolean) parameters[0]).booleanValue();   
	    setUnderAttack(b);
	    break;
	case DECREASE_FUEL_LEVEL:
	    d =((Double) parameters[0]).doubleValue();
	    decreaseFuelLevel(d);
	    break;
	case SET_RELOAD_PROPERTY:
	    s =(String) parameters[0];
	    setReloadProperty(s);
	    break;
	case DECREASE_AMMUNITION_LEVEL :
	    n = ((Integer) parameters[0]).intValue();
	    decreaseAmmunitionLevel(n);
	    break;
	case SET_REGAINING_STRENGTH :
	    b =((Boolean) parameters[0]).booleanValue();   
	    setRegainingStrength(b);
	    break;
	case SET_REGAIN_STRENGTH_PROPERTY:
	    s =(String) parameters[0];
	    setRegainStrengthProperty(s);
	    break;
	case SEND_UNIT_MESSAGE:
	    s =(String) parameters[0];
	    sendUnitMessage(s);
	    break;
	case PERFORM_RETREAT_TO :
	    x = ((Double) parameters[0]).doubleValue();
	    y = ((Double) parameters[1]).doubleValue();
	    performRetreatTo(x,y);
	    break;

		   
	default:
	    // Debug info
	    System.out.println("WarUnit:playReplayData(int, Object[]) method " + method + " not handled!");
	    break;
	} 
    }

}

