package gecco.server.replaymanager;

import java.io.ObjectInputStream;
import java.io.FileInputStream;
import gecco.util.SignalRunnable;
import gecco.server.core.ReferenceHolder;
import gecco.server.core.Unit;
import gecco.server.clock.Clock;
import gecco.util.SignalThread;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.EOFException;
import gecco.server.automaton.Automaton;
import java.text.DecimalFormat;


/**
 * ReplayManager.java
 *
 *
 * Created: Tue May 20 21:08:18 2003
 *
 * @author <a href="mailto:hans.sandstrom@fhs.mil.se">Hans Sandstrm</a>
 * @version
 */

public class ReplayManager implements SignalRunnable  {

    ObjectInputStream replayData;

    public ReplayManager(ObjectInputStream ois){
	replayData = ois;
    }

    public ReplayManager(String s){
	this();
	try {
	    replayData = new ObjectInputStream(new FileInputStream(s));
	} catch (FileNotFoundException e) {
	    ReferenceHolder.err.println("ReplayManager: Replay data file '" + s + " not found, exiting!");
	    e.printStackTrace();
	    System.exit(1);
	} catch (IOException e){
	    ReferenceHolder.err.println("ReplayManager: IOException, exiting!");
	    e.printStackTrace();
	    System.exit(1);
	}
	// Debug info
	System.out.println("ReplayManager: ReplayManager initiated");
    }

    private ReplayManager(){
    }


    // implementation of java.lang.Runnable interface

    /**
     *
     */
    public void run(){
	// Debug info
	System.out.println("ReplayManager: thread started");
	
	double timeStamp = -1;
	try {
	    timeStamp = replayData.readDouble();
	} catch (IOException e) {
	    ReferenceHolder.err.println("ReplayManager: Can't read any replay data, exiting!");
	    ReferenceHolder.err.println(e.getMessage());
	    e.printStackTrace();
	    System.exit(1);
	}
	
        while (true) {
	    
	    // Check for signals.
	    switch (SignalThread.signalPending()) {
	    case SIGKILL:
	    case SIGQUIT:
		// Stop executing.
		return;
	    case SIGPAUSE:
		SignalThread.pauseHandler();
		SignalThread.clearSignal();
		break;
	    default:
		// Do nothing.
		SignalThread.clearSignal();
		break;
	    }
	    
	    double gameTime = ReferenceHolder.getClock().gameTimeNow();
	    
	    if (timeStamp > gameTime) {
		try {
		    // Sleep until next data should be replayed
		    ReferenceHolder.getClock().gameSleep(timeStamp - gameTime);
		} catch (InterruptedException e) {
		    // Time to wakeup?
		}
	    }
	    else {
		// Time to replay data
		try {
		    int type = replayData.readInt();
		    if (type == ReplayReady.UNIT) {
			int unitHandle = replayData.readInt();
			int method = replayData.readInt();
			Object[] parameters = (Object[]) replayData.readObject();
			Unit unit = ReferenceHolder.getUnitManager().getUnit(unitHandle);
			ReplayReady rr = ((ReplayReady) unit);
			// Save old position and reset updated checks
			double oldx = unit.getX();
			double oldy = unit.getY();
			unit.resetUpdatedActions();
			unit.resetUpdatedProperties();
			// Replay
			rr.playReplayData(method, parameters);
			// Check if things has changed...
			if (unit.isDestroyed()) {
			    ReferenceHolder.getUnitManager().deleteUnit(unit, oldx, oldy);
			    ReferenceHolder.getVisibilityFilter().deleteUnit(unit, oldx, oldy);
			} else {
			    ReferenceHolder.getActionProcessor().publicCheckChangedPosition(oldx, oldy, unit);
			    ReferenceHolder.getActionProcessor().publicCheckChangedProperties(unit);
			    ReferenceHolder.getActionProcessor().publicCheckActionBlocklist(unit);
			}
		    }
		    else if (type == ReplayReady.AUTOMATON) {
			int posX = replayData.readInt();
			int posY = replayData.readInt();
			int function = replayData.readInt();
			Object[] parameters = (Object[]) replayData.readObject();
			Automaton automaton = ReferenceHolder.getAutomatonMap().getAutomaton(posX, posY);
			ReplayReady rr = ((ReplayReady) automaton);
			rr.playReplayData(function, parameters);
		    }
		    else if (type == ReplayReady.MOVING_ACTION) {
			int actionHandle = replayData.readInt();
			int unitHandle = replayData.readInt();
			boolean moving = replayData.readBoolean();
			// Debug info
// 			DecimalFormat df = new DecimalFormat(".00");
// 			System.out.println("ReplayManager: Action " + actionHandle + 
// 					   " reports unit " + unitHandle + " moving = " + moving + 
// 					   " at " + df.format(timeStamp));

		    }
		    // when should next data be replayed?
		    timeStamp = replayData.readDouble();
		} catch (EOFException e) {
		    ReferenceHolder.getClock().freeze(true);
		    ReferenceHolder.out.println("ReplayManager: End of replay data.");
		    return;
		} catch (Exception e){
		    ReferenceHolder.err.println("ReplayManager: Something whent wrong reading replay data, exiting!");
		    ReferenceHolder.err.println(e.getMessage());
		    e.printStackTrace();
		    System.exit(1);
		}
	    }
	}
    }
    
}// ReplayManager
